;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Sendfile Egg for the Chicken Scheme system.
;;              
;; This eggs provides a capability to utilize
;; the sendfile system-call. However it is
;; not a mere wrapper to call this function if
;; available, but rather its semantics may be stated as:
;;
;; "Send the file as fast as possible to its destination."
;;
;; Please report bugs to <http://trac.callcc.org/>
;;
;; Copyright (c) 2007 David Krentzlin 
;;
;;
;; Thanks to Peter Bex for supplying patches and testing it on NetBSD
;;
;; Permission is hereby granted, free of charge, to any person obtaining a
;; copy of this software and associated documentation files (the "Software"),
;; to deal in the Software without restriction, including without limitation
;; the rights to use, copy, modify, merge, publish, distribute, sublicense,
;; and/or sell copies of the Software, and to permit persons to whom the
;; Software is furnished to do so, subject to the following conditions:
;;
;; The above copyright notice and this permission notice shall be included
;; in all copies or substantial portions of the Software.
;; 
;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
;; OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
;; ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
;; OTHER DEALINGS IN THE SOFTWARE.
;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;




;;sendfile attempts to send a file from the sourcefile out to a tcp-socket
;;as fast as possible. On systems that provide sendfile(2) syscalls
;;this syscall will be used if apropriate on other systems memory-mapped write
;;will be used to emulate this. And if even this fails a normal system-write
;;will be used to send the data


;;TODO
;;add -D_FILE_OFFSET_BITS=64 -D_LARGEFILE_SOURCE -D_LARGE_FILES to sendfile.setup

(declare
 (usual-integrations)
 (uses extras scheduler)
 (no-procedure-checks-for-usual-bindings)
 (export
  *sendfile:last-selected-implementation*
  sendfile:read-write-buffer-size
  sendfile:write-timeout
  sendfile:implementation-selector
  sendfile:mmapped
  sendfile:sendfile
  sendfile:read-write-loop
  sendfile:os-dep:sendfile-available?
  sendfile:os-dep:mmap-available?
  sendfile))
(declare
 (foreign-declare #<<EOL
#include "os-dep.h"
#define C_write(fd, b, n)   C_fix(write(C_unfix(fd), C_data_pointer(b), C_unfix(n)))
EOL
))

(use posix lolevel srfi-4)

;;what type of system do we have?
(define-foreign-variable %sendfile:word-size int "C_WORD_SIZE")


;;is the syscall present?
(define-foreign-variable %sendfile:have-native-sendfile int "HAVE_SENDFILE")

;;do we have mmap?
(define-foreign-variable %sendfile:have-mmap int "HAVE_MMAP")
(define-foreign-variable %sendfile:have-madvise int "HAVE_MADVISE")
(define-foreign-variable %sendfile:madvise-sequential int "MADV_SEQUENTIAL")


(define-foreign-variable %sendfile:bufsize int "BUFSIZ")


;;the sendfile implementation
;;the off_t parameter is a flonum
;;as file-size returns a flonum on big files
(define %sendfile:sendfile-implementation
  (foreign-lambda double "sendfile_implementation" int int double unsigned-integer))


(define sendfile:os-dep:sendfile-available? (= %sendfile:have-native-sendfile 1))
(define sendfile:os-dep:mmap-available? (= %sendfile:have-mmap 1))

;;the buffer used in read write loops
;;the client may adjust this to meet its need
(define sendfile:read-write-buffer-size (make-parameter %sendfile:bufsize))


;;biggest unsigned integer
(define *sendfile:chunk-size* (expt 2 (- %sendfile:word-size 1)))


(define (sendfile:madvise buff len behav)
  (if (= %sendfile:have-madvise 1)
      ((foreign-lambda int "madvise" (pointer char) unsigned-integer int) buff len behav)))
      

;;compute the next chunk to send
;;if total-length - offset is bigger than the biggest
;;integer on the system we send the biggest integer
;;otherwise we send the difference
(define (sendfile:next-chunk-size len offset)
    (if (> (- len offset) *sendfile:chunk-size*)
        *sendfile:chunk-size*
        (- len offset)))


;;posix mmapped send
;;as we're working with non-blocking IO
;;EAGAIN is equal to EWOULDBLOCK
(define (sendfile:mmapped src dst len)
  (unless sendfile:os-dep:mmap-available?
    (sendfile:error "mmap is not available on this system"))

  (define sys-write (foreign-lambda integer "write" integer c-pointer unsigned-integer))

  (define (send-chunk size ptr)
    (sendfile:madvise ptr size %sendfile:madvise-sequential)
    
    (let loop ((left size) (work-ptr (pointer-offset ptr 0)))
      (cond
       ((= left 0) #t)
       (else
        (let ((result (sys-write dst work-ptr left)))
          (cond
           ((and (negative? result) (= errno/again (##sys#update-errno)))
            (loop left work-ptr))
           ((negative? result) #f)
           (else
            (loop (- left result) (pointer-offset work-ptr result)))))))))

  (let loop ((offset 0.0))
    (cond
     ((= offset len) len)
     (else
      (let* ((next-chunk (sendfile:next-chunk-size len offset))
             (mem-file (map-file-to-memory #f next-chunk prot/read (bitwise-ior map/shared map/file) src offset)))
        (unless (send-chunk next-chunk (memory-mapped-file-pointer mem-file))
          (unmap-file-from-memory mem-file)
          (sendfile:error "write failed"))
        (unmap-file-from-memory mem-file)
        (loop (+ offset next-chunk)))))))
             
                       
;;the interface to the sendfile-implementation
;;as we're working with non-blocking IO
;;EAGAIN is equal to EWOULDBLOCK
;;MAKE THE WRITE LOOP AN INNER LOOP
;;SO THAT ONLY THE ACTUAL SYSCALL IS REDONE AND NOT THE CALCULATION
(define (sendfile:sendfile src dst len)
  (unless sendfile:os-dep:sendfile-available?
    (sendfile:error "sendfile is not available on this system"))
  (let loop ((offset 0.0))
    (cond
     ((= offset len)  len)
     (else
      (let* ((next-chunk (sendfile:next-chunk-size len offset))
             (new-offset (%sendfile:sendfile-implementation src dst offset next-chunk)))
        (cond
         ((negative? new-offset)
          (sendfile:error "sendfile failed"))
         (else
          (loop new-offset))))))))
        

      
;;as we're working with non-blocking IO
;;EAGAIN is equal to EWOULDBLOCK

(define (sendfile:read-write-loop src dst len)
  (let* ((buff-size (sendfile:read-write-buffer-size))
         (buffer (make-string buff-size)))
        
    
    (define (write-bytes size)
      (let loop ((left size))
        (let ((written-bytes (##core#inline "C_write" dst buffer left)))
          (cond
           ((= 0 left) #t)
           ((and (negative? written-bytes) (= errno/again (##sys#update-errno)))
            (loop left)) ;;try again
           ((negative? written-bytes)
            (sendfile:error "write failed"))
           (else
            (loop (- left written-bytes)))))))

    (let loop ((n len))
      (cond
       ((not (positive? n)) len)
       (else
        (let* ((shall-read (if (> n buff-size) buff-size (inexact->exact n)))
               (read-bytes (cadr (file-read src shall-read buffer))))
          (write-bytes read-bytes)
          (loop (- n read-bytes))))))))
           
         
;;=============================== HIGHER LEVEL INTERFACE =======================================================

(define *sendfile:last-selected-implementation* #f)

(define (sendfile:default-selector len)
  (cond
   ((< len (* 1024 1024)) ;;below mb we don't bother memory-mapping or sendfile
    (set!  *sendfile:last-selected-implementation* 'read-write-loop)
    sendfile:read-write-loop)
   (sendfile:os-dep:sendfile-available?
    (set!  *sendfile:last-selected-implementation* 'sendfile)
    sendfile:sendfile)
   (sendfile:os-dep:mmap-available?
    (set!  *sendfile:last-selected-implementation* 'mmapped)
    sendfile:mmapped)
   (else
    (set!  *sendfile:last-selected-implementation* 'read-write-loop)
    sendfile:read-write-loop)))
   
 
(define sendfile:implementation-selector (make-parameter sendfile:default-selector))


(define (sendfile:->filenum obj)
  (if (port? obj)
      (port->fileno obj)
      (if (number? obj)
          obj
          (sendfile:error "supplied type is neither port nor filedescriptor"))))
 
(define (sendfile:error msg)
  (abort (make-property-condition 'exn
                                  'message
                                  msg)))

(define (sendfile src dst #!key (force 'nothing))
  (let* ((src (sendfile:->filenum src))
         (dst (sendfile:->filenum dst))
         (len (file-size src)))
     (case force
       ((sendfile) (sendfile:sendfile src dst len))
       ((mmapped) (sendfile:mmapped src dst len))
       ((read-write) (sendfile:read-write-loop src dst len))
       (else
        (((sendfile:implementation-selector) len) src dst len)))))

  

                          
                          


      



                             
                         
  


                           




  
    